;; -*- mode: Scheme; -*-
;; This file is part of zxcvbn for CHICKEN
;; Copyright (c) 2022 by Thomas Chust. All rights reserved.
;;
;; Permission is hereby granted, free of charge, to any person
;; obtaining a copy of this software and associated documentation
;; files (the Software), to deal in the Software without restriction,
;; including without limitation the rights to use, copy, modify,
;; merge, publish, distribute, sublicense, and/or sell copies of the
;; Software, and to permit persons to whom the Software is furnished
;; to do so, subject to the following conditions:
;;
;; The above copyright notice and this permission notice shall be
;; included in all copies or substantial portions of the Software.
;;
;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
;; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
;; BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
;; ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
;; CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
;; SOFTWARE.

(module zxcvbn
  (zxcvbn-match zxcvbn-match*)
  (import
   scheme
   (chicken base)
   (chicken foreign)
   (chicken memory)
   (only (chicken bitwise) bitwise-and bit->boolean)
   (only (chicken string) string-intersperse))

(foreign-declare "#include \"zxcvbn.h\"")

(define %zcvbn-match
  (foreign-lambda* double ([nonnull-c-string passwd]
                           [pointer-vector userdict]
                           [(c-pointer (c-pointer "ZxcMatch_t")) info])
    "C_return(ZxcvbnMatch(passwd, (const char **)userdict, info));"))

(define (string-list->pointer-vector strs)
  (if (or (not strs) (null? strs))
    (values #f void)
    (let loop ([count 1] [size 0] [ss (or strs '())])
      (if (null? ss)
        (let ([data (and (positive? size) (allocate size))]
              [ptrs (make-pointer-vector count #f)])
          (let loop ([i 0] [p data] [ss (or strs '())])
            (if (null? ss)
              (values
                ptrs
                (cut free data))
              (let* ([s (car ss)] [n (string-length s)])
                (move-memory! s p n)
                (pointer-u8-set! (pointer+ p n) 0)
                (pointer-vector-set! ptrs i p)
                (loop (add1 i) (pointer+ p (add1 n)) (cdr ss))))))
        (loop (add1 count) (+ size (string-length (car ss)) 1) (cdr ss))))))

(define %zxcvbn-free-info
  (foreign-lambda
    void "ZxcvbnFreeInfo"
    (c-pointer "ZxcMatch_t")))

(define %zxc-match-begin
  (foreign-lambda* int ([(nonnull-c-pointer "ZxcMatch_t") info])
    "C_return(info->Begin);"))

(define %zxc-match-length
  (foreign-lambda* int ([(nonnull-c-pointer "ZxcMatch_t") info])
    "C_return(info->Length);"))

(define %zxc-match-entropy
  (foreign-lambda* double ([(nonnull-c-pointer "ZxcMatch_t") info])
    "C_return(info->Entrpy);"))

(define %zxc-match-multipart-entropy
  (foreign-lambda* double ([(nonnull-c-pointer "ZxcMatch_t") info])
    "C_return(info->MltEnpy);"))

(define (%zxc-match-type info)
  (let ([typ ((foreign-lambda* int ([(nonnull-c-pointer "ZxcMatch_t") info])
                "C_return(info->Type);")
              info)])
    (values
      (case (bitwise-and typ #b11111)
        [(1)  'brute]
        [(2)  'dict]
        [(3)  'dict/leet]
        [(4)  'user]
        [(5)  'user/leet]
        [(6)  'repeats]
        [(7)  'sequence]
        [(8)  'spatial]
        [(9)  'date]
        [(10) 'year]
        [else #f])
      (bit->boolean typ 5))))

(define %zxc-match-next
  (foreign-lambda* (c-pointer "ZxcMatch_t") ([(nonnull-c-pointer "ZxcMatch_t") info])
    "C_return(info->Next);"))

(define (zxc-match-info->list info)
  (if (or (not info) (zero? (pointer->address info)))
    '()
    (cons
      (let*-values ([(typ mm?) (%zxc-match-type info)]
                    [(it)
                     (list
                       typ
                       (%zxc-match-begin info) (%zxc-match-length info)
                       (%zxc-match-entropy info) (%zxc-match-multipart-entropy info))])
        (if mm? (cons '* it) it))
      (zxc-match-info->list (%zxc-match-next info)))))

(define (zxcvbn-match passwd #!optional userdict)
  (let*-values ([(userdict release!) (string-list->pointer-vector userdict)]
                [(entropy) (%zcvbn-match passwd userdict #f)])
    (release!)
    entropy))

(define (zxcvbn-match* passwd #!optional userdict)
  (let-location ([info (c-pointer "ZxcMatch_t") (address->pointer 0)])
    (let*-values ([(userdict release!) (string-list->pointer-vector userdict)]
                  [(entropy) (%zcvbn-match passwd userdict #$info)])
      (release!)
      (let ([matches (zxc-match-info->list info)])
        (%zxcvbn-free-info info)
        (values entropy matches)))))

)
;; vim:set ai et ts=4 sts=2 sw=2 ft=scheme: ;;
